#!/usr/bin/env python3


import pickle
import os
import sys


mat = {}
header = []

if not os.path.isfile(sys.argv[1]+'.pkl'):
    with open(sys.argv[1]) as f:
        for row in f:
            if not row.startswith('#'):
                row_clean = row.strip().split()

                if not header:
                    header = row_clean
                else:
                    aa = row_clean[0]

                    for i, s in enumerate(row_clean[1:]):
                        mat[(header[i], aa)] = float(s)

    with open(sys.argv[1]+'.pkl', 'wb') as f:
        pickle.dump(mat, f, protocol=pickle.HIGHEST_PROTOCOL)



#!/usr/bin/env python


__author__ = ('Francesco Asnicar (f.asnicar@unitn.it), '
              'Francesco Beghini (francesco.beghini@unitn.it), '
              'Claudia Mengoni (claudia.mengoni@studenti.unitn.it), '
              'Mattia Bolzan (mattia.bolzan@unitn.it), '
              'Nicola Segata (nicola.segata@unitn.it)')
__version__ = '3.0.23'
__date__ = '27 November 2020'


import sys
import bz2
import glob
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import os
import argparse as ap
from urllib.request import urlretrieve
from urllib.parse import urlencode
from urllib.request import Request
from urllib.request import urlopen
import time


if sys.version_info[0] < 3:
    raise Exception("PhyloPhlAn {} requires Python 3, your current Python version is {}.{}.{}"
                    .format(__version__, sys.version_info[0], sys.version_info[1], sys.version_info[2]))

DB_TYPE_CHOICES = ['n', 'a']
GENOME_EXTENSION = '.fna'
PROTEOME_EXTENSION = '.faa'
#DOWNLOAD_URL = "https://www.dropbox.com/s/jrituki21vx6p30/taxa2core.txt?dl=1"
DOWNLOAD_URL = "http://cmprod1.cibio.unitn.it/databases/PhyloPhlAn/taxa2core.txt"


def info(s, init_new_line=False, exit=False, exit_value=0):
    if init_new_line:
        sys.stdout.write('\n')

    sys.stdout.write('{}'.format(s))
    sys.stdout.flush()

    if exit:
        sys.exit(exit_value)


def error(s, init_new_line=False, exit=False, exit_value=1):
    if init_new_line:
        sys.stderr.write('\n')

    sys.stderr.write('[e] {}\n'.format(s))
    sys.stderr.flush()

    if exit:
        sys.exit(exit_value)


def read_params():
    p = ap.ArgumentParser(description=("The phylophlan_setup_database.py script can be used to either format an input folder or "
                                       "multi-fasta file to be used as database in phylophlan.py, or automatically download a "
                                       "pre-identified set of core UniRef90 proteins for the taxonomic label of a given species"),
                          formatter_class=ap.ArgumentDefaultsHelpFormatter)

    group = p.add_mutually_exclusive_group()
    group.add_argument('-i', '--input', type=str,
                       help=("Specify the path to either the folder containing the marker files or the file of markers, in "
                             "(multi-)fasta format"))
    group.add_argument('-g', '--get_core_proteins', type=str, default=None,
                       help=('Specify the taxonomic label for which download the set of core proteins. The label must represent a '
                             'species: "--get_core_proteins s__Escherichia_coli"'))

    p.add_argument('--database_update', action='store_true', default=False, help="Update the databases file")
    p.add_argument('-o', '--output', type=str, default=None, help="Specify path to the output folder where to save the database")
    p.add_argument('-d', '--db_name', type=str, help="Specify the name of the output database")
    p.add_argument('-e', '--input_extension', type=str, default=None,
                   help="Specify the extension of the input file(s) specified via -i/--input")
    p.add_argument('-t', '--db_type', default=None, choices=DB_TYPE_CHOICES,
                   help='Specify the type of the database, where "n" stands for nucleotides and "a" for amino acids')
    p.add_argument('-x', '--output_extension', type=str, default=None, help="Set the database output extension")
    p.add_argument('--overwrite', action='store_true', default=False, help="If specified and the output file exists it will be overwritten")
    p.add_argument('--citation', action='version',
                   version=('Asnicar, F., Thomas, A.M., Beghini, F. et al. '
                            'Precise phylogenetic analysis of microbial isolates and genomes from metagenomes using PhyloPhlAn 3.0. '
                            'Nat Commun 11, 2500 (2020). '
                            'https://doi.org/10.1038/s41467-020-16366-7'),
                   help="Show citation")
    p.add_argument('--verbose', action='store_true', default=False, help="Prints more stuff")
    p.add_argument('-v', '--version', action='version',
                   version='phylophlan_setup_database.py version {} ({})'.format(__version__, __date__),
                   help="Prints the current phylophlan_setup_database.py version and exit")
    return p.parse_args()


def check_params(args, verbose=False):
    if args.database_update:
        database_update(update=args.database_update, verbose=args.verbose)
        args.database_update = False

    if not args.input and not args.get_core_proteins:
        error('either -i/--input or -g/--get_core_proteins must be specified', init_new_line=True, exit=True)

    if args.input:
        if (not os.path.isdir(args.input)) and (not os.path.isfile(args.input)):
            error('input must be either a folder or a file', exit=True)

        if not args.db_name:
            error('-d/--db_name must be specified', exit=True)

        if os.path.isdir(args.input):
            if not args.input_extension:
                error('input is a folder, hence --input_extension must be specified', exit=True)

    if args.get_core_proteins:
        # not handling keys different than species (Tue 19 Jun 2018)
        if args.get_core_proteins[:3] != 's__':
            error('The taxonomic label provided "{}" does not starts with "s__"'.format(args.get_core_proteins),
                  exit=True)

        if not args.output:
            output_tmp = os.path.dirname(args.get_core_proteins)

            if output_tmp:
                args.output = output_tmp
            else:
                args.output = args.get_core_proteins

            if verbose:
                info('Setting output folder "{}"\n'.format(args.output))
        elif os.path.isdir(args.output) and (not args.output.endswith(args.get_core_proteins)):
            args.output = os.path.join(args.output, args.get_core_proteins)

            if verbose:
                info('Setting output folder "{}"\n'.format(args.output))

        args.input = args.output
        args.input_extension = PROTEOME_EXTENSION
        args.output_extension = PROTEOME_EXTENSION

        if not args.db_name:
            args.db_name = args.get_core_proteins

            if verbose:
                info('Setting db_name to "{}"\n'.format(args.db_name))

    if not args.db_type:
        if not args.output_extension:
            error('either -t/--db_type or -x/--output_extension must be specified', exit=True)
        else:
            if not args.output_extension.startswith('.'):
                args.output_extension = '.' + args.output_extension

            if args.output_extension.endswith('.'):
                args.output_extension = args.output_extension[:-1]
    else:
        if not args.output_extension:
            if 'n' == args.db_type:
                args.output_extension = GENOME_EXTENSION
            if 'a' == args.db_type:
                args.output_extension = PROTEOME_EXTENSION

            if verbose:
                info('Setting output extension to "{}"\n'.format(args.output_extension))
        else:
            error("both -t/--db_type and -x/--output_extension were specified, don't know which one to use!",
                  exit=True)

    if not args.output:
        args.output = args.input if os.path.isdir(args.input) else os.path.dirname(args.input)

        if verbose:
            info('Output folder not specified, setting to "{}"\n'.format(args.output))

    if verbose:
        info('Arguments: {}\n'.format(vars(args)))


def database_update(update=False, verbose=False):
    taxa2core_file_latest = None
    taxa2core_file = os.path.basename(DOWNLOAD_URL).replace('?dl=1', '')
    download(DOWNLOAD_URL, taxa2core_file, overwrite=update, verbose=verbose)

    with open(taxa2core_file) as f:
        for r in f:
            if not r.startswith('#'):
                taxa2core_file_latest, taxa2core_file_latest_url = r.strip().split('\t')
                break  # file should contains only one line, i.e., the name of the latest taxa2core file

    taxa2core_file_latest = taxa2core_file_latest.replace('?dl=1', '')
    download(taxa2core_file_latest_url, taxa2core_file_latest, overwrite=update, verbose=verbose)

    return taxa2core_file_latest


def byte_to_megabyte(byte):
    """
    Convert byte value to megabyte
    """

    return (byte / 1048576)


class ReportHook():

    def __init__(self):
        self.start_time = time.time()

    def report(self, blocknum, block_size, total_size):
        """
        Print download progress message
        """

        if blocknum == 0:
            self.start_time = time.time()

            if total_size > 0:
                info("Downloading file of size: {:.2f} MB\n".format(byte_to_megabyte(total_size)))
        else:
            total_downloaded = blocknum * block_size
            status = "{:3.2f} MB ".format(byte_to_megabyte(total_downloaded))

            if total_size > 0:
                percent_downloaded = total_downloaded * 100.0 / total_size
                # use carriage return plus sys.stderr to overwrite stderr
                download_rate = total_downloaded / (time.time() - self.start_time)
                estimated_time = (total_size - total_downloaded) / download_rate
                estimated_minutes = int(estimated_time / 60.0)
                estimated_seconds = estimated_time - estimated_minutes * 60.0
                status += ("{:3.2f} %  {:5.2f} MB/sec {:2.0f} min {:2.0f} sec "
                           .format(percent_downloaded, byte_to_megabyte(download_rate),
                                   estimated_minutes, estimated_seconds))

            status += "        \r"
            info(status)


def download(url, download_file, overwrite=False, verbose=False):
    """
    Download a file from a url
    """

    if overwrite or not os.path.isfile(download_file):
        try:
            if verbose:
                info('Downloading "{}" to "{}"\n'.format(url, download_file))

            urlretrieve(url, download_file, reporthook=ReportHook().report)
            info('\n')
        except EnvironmentError:
            error('unable to download "{}"'.format(url))
    elif verbose:
        info('File "{}" present\n'.format(download_file))


def create_folder(output, verbose=False):
    if not os.path.exists(output):
        if verbose:
            info('Creating output folder "{}"\n'.format(output))

        os.mkdir(output, mode=0o775)
    elif verbose:
        info('Output "{}" folder present\n'.format(output))


def get_core_proteins(taxa2core_file, taxa_label, output, output_extension, verbose=False):
    core_proteins = {}
    url = None
    retry2download = []
    not_mapped = []
    not_mapped_again = []

    # taxa2core format
    #
    #   # taxid   taxonomy      UniRefURL       core_list
    #   12345     k__|...|s__   http://..{}..   UP123;UP456;...

    for r in bz2.open(taxa2core_file, 'rt'):
        if r.startswith('#'):
            continue

        r_clean = r.strip().split('\t')

        if taxa_label in r_clean[1]:
            if taxa_label == r_clean[1].split('|')[-1]:
                url = r_clean[2]
                core_proteins[r_clean[1]] = r_clean[3].split(';')
                break
            else:
                core_proteins[r_clean[1]] = None

    if not len(core_proteins):
        error('no entry found for "{}", please check the taxonomic label provided'.format(taxa_label),
              exit=True)
    elif len([k for k, v in core_proteins.items() if v is not None]) > 1:
        error('{} entries found for "{}":\n{}    please check the taxonomic label provided'
              .format(len(core_proteins), taxa_label, '    - {}\n'.join(core_proteins.keys())), exit=True)

    for lbl, core_prots in core_proteins.items():
        if core_prots is None:
            continue

        if verbose:
            info('Downloading {} core proteins for {}\n'.format(len(core_prots), lbl))

        for core_prot in core_prots:
            local_prot = os.path.join(output, core_prot + output_extension)
            download(url.format(core_prot), local_prot, verbose=verbose)

            if not os.path.exists(local_prot):
                retry2download.append(core_prot)

    # try to re-map the ids in case "not mapped" store in not_mapped
    if retry2download:
        if verbose:
            info("Re-trying to download {} core proteins that just failed, please wait as it might take some time\n"
                 .format(len(retry2download)))
        idmapping_url = 'https://www.uniprot.org/uploadlists/'
        contact = "phylophlan@cibiocm.com"
        params = {'from': 'ACC+ID',
                  'to': 'NF90',
                  'format': 'tab',  # or 'list' for only converted clusters
                  'query': ' '.join(retry2download)}
        data = urlencode(params).encode('utf-8')
        request = Request(idmapping_url, data, headers={'User-Agent': 'Python {}'.format(contact)})

        try:
            response = urlopen(request, data)
            uniprotkb2uniref90 = [line.decode().strip().split('\t')[:2] for line in response.readlines()][1:]  # skip ['From', 'To']
        except Exception:
            error('unable convert UniProtKB ID to UniRef90 ID')

        for uniref90_id in (x[1].split('_')[-1] for x in uniprotkb2uniref90[1:]):
            local_prot = os.path.join(output, uniref90_id + output_extension)
            download(url.format(uniref90_id), local_prot, verbose=verbose)

            if not os.path.exists(local_prot):
                not_mapped.append(uniref90_id)

        if len(uniprotkb2uniref90) != len(retry2download):
            # probably deleted proteins in the Uniprot versions, try to download their latest version in any case
            for ur90 in set([ur90 for ur90, _ in uniprotkb2uniref90]) - set(retry2download):
                local_prot = os.path.join(output, ur90 + output_extension)
                download('https://www.uniprot.org/uniprot/{}.fasta?version=*'.format(ur90), local_prot, verbose=verbose)

                if not os.path.exists(local_prot):
                    not_mapped_again.append(core_prot)

        # really don't know what else to try... I'm sorry!
        if not_mapped_again:
            nd_out = os.path.join(output, taxa_label + '_core_proteins_not_mapped.txt')

            if verbose:
                info('There are {} core proteins that could not be downloaded, writing thier IDs to "{}"\n'
                     .format(len(not_mapped_again), nd_out))

            with open(nd_out, 'w') as f:
                f.write('\n'.join(not_mapped_again) + '\n')


def create_database(db_name, inputt, input_ext, output, overwrite, verbose=False):
    seqs = []

    if os.path.exists(output) and (not overwrite):
        error('output file exists and --overwrite not specified', exit=True)

    if os.path.isdir(inputt):
        for marker in glob.iglob(os.path.join(inputt, '*' + input_ext + '*')):
            seqs += [SeqRecord(record.seq,
                               id='_'.join([db_name.replace('_', '-').replace(',', '-').replace(':', '').replace('|', '-'),
                                            record.id.replace('_', '-').replace(',', '-').replace(':', '').replace('|', '-'),
                                            str(count)]),
                               description='')
                     for count, record in enumerate(SeqIO.parse(bz2.open(marker, 'rt') if marker.endswith('.bz2')
                                                                else open(marker), "fasta"))]
    else:
        seqs = [SeqRecord(record.seq,
                          id='_'.join([db_name.replace('_', '-').replace(',', '-').replace(':', '').replace('|', '-'),
                                       record.id.replace('_', '-').replace(',', '-').replace(':', '').replace('|', '-'),
                                       str(count)]),
                          description='')
                for count, record in enumerate(SeqIO.parse(bz2.open(inputt, 'rt') if inputt.endswith('.bz2')
                                                           else open(inputt), "fasta"))]

    if not seqs:
        error('no sequences found, make sure the input folder/file provided is not empty', exit=True)

    if verbose:
        info('Writing output database "{}"\n'.format(output))

    with open(output, 'w') as f:
        SeqIO.write(seqs, f, "fasta")


def phylophlan_setup_database():
    args = read_params()

    if args.verbose:
        info('phylophlan_setup_database.py version {} ({})\n'.format(__version__, __date__))
        info('Command line: {}\n\n'.format(' '.join(sys.argv)), init_new_line=True)

    check_params(args, verbose=args.verbose)
    create_folder(args.output, verbose=args.verbose)

    if args.get_core_proteins:
        taxa2core_file_latest = database_update(update=args.database_update, verbose=args.verbose)
        get_core_proteins(taxa2core_file_latest, args.get_core_proteins, args.output, args.output_extension, verbose=args.verbose)

    create_database(args.db_name, args.input, args.input_extension, os.path.join(args.output, args.db_name + args.output_extension),
                    args.overwrite, verbose=args.verbose)


if __name__ == '__main__':
    t0 = time.time()
    phylophlan_setup_database()
    t1 = time.time()
    info('Total elapsed time {}s\n'.format(int(t1 - t0)), init_new_line=True)
    sys.exit(0)



#!/usr/bin/env python


__author__ = ('Francesco Asnicar (f.asnicar@unitn.it), '
              'Claudia Mengoni (claudia.mengoni@studenti.unitn.it)')
__version__ = '3.0.10'
__date__ = '28 May 2020'


import argparse as ap
import os
import sys
import time
import datetime
import pandas as pd
import itertools
from Bio import Phylo


if sys.version_info[0] < 3:
    raise Exception("PhyloPhlAn {} requires Python 3, your current Python version is {}.{}.{}"
                    .format(__version__, sys.version_info[0], sys.version_info[1], sys.version_info[2]))

NEWICK = 'newick'
TREE_TYPES = ['newick', 'nexus', 'phyloxml', 'cdao', 'nexml']
PHYLOGENETIC_THR = 0.05
MUT_PERCENTAGE_THR = 0.05
OUTPUT_EXTENSIONS = {';': '.txt',
                     ',': '.csv',
                     '\t': '.tsv'}


def info(s, init_new_line=False, exit=False, exit_value=0):
    if init_new_line:
        sys.stdout.write('\n')

    sys.stdout.write('{}'.format(s))
    sys.stdout.flush()

    if exit:
        sys.exit(exit_value)


def error(s, init_new_line=False, exit=False, exit_value=1):
    if init_new_line:
        sys.stderr.write('\n')

    sys.stderr.write('[e] {}\n'.format(s))
    sys.stderr.flush()

    if exit:
        sys.exit(exit_value)


def read_params():
    p = ap.ArgumentParser(description=("The phylophlan_strain_finder.py script analyzes the phylogeny and the mutation rates table "
                                       "generated from the phylophlan.py script and returns sub-trees representing the same strain, "
                                       "according to both a phylogenetic threshold (computed on the normalized pairwise phylogenetic "
                                       "distances) and a mutation rate threshold (computed on the aligned sequences of the markers used "
                                       "in the phylogenetic analysis)"),
                          formatter_class=ap.ArgumentDefaultsHelpFormatter)

    p.add_argument('-i', '--input', type=str, required=True,
                   help='Specify the file of the phylogenetic tree as generated from phylophlan.py')
    p.add_argument('-m', '--mutation_rates', type=str, required=True,
                   help='Specify the file of the mutation rates as generated from phylophlan.py')
    p.add_argument('--phylo_thr', type=float, default=PHYLOGENETIC_THR,
                   help='Maximum phylogenetic distance threshold for every pair of nodes in the same subtree (inclusive)')
    p.add_argument('--mutrate_thr', type=float, default=MUT_PERCENTAGE_THR,
                   help='Maximum mutation rate ratio for every pair of nodes in the same subtree (inclusive)')
    p.add_argument('--tree_format', choices=TREE_TYPES, default=NEWICK, help='Specify the format of the input tree')
    p.add_argument('-o', '--output', type=str, default=None, help='Specify the output filename, if not specified will be stdout')
    p.add_argument('--overwrite', action='store_true', default=False, help='Overwrite the output file if exists')
    p.add_argument('-s', '--separator', type=str, default='\t', choices=OUTPUT_EXTENSIONS.keys(),
                   help='Specify the separator to use in the output')
    p.add_argument('--citation', action='version',
                   version=('Asnicar, F., Thomas, A.M., Beghini, F. et al. '
                            'Precise phylogenetic analysis of microbial isolates and genomes from metagenomes using PhyloPhlAn 3.0. '
                            'Nat Commun 11, 2500 (2020). '
                            'https://doi.org/10.1038/s41467-020-16366-7'),
                   help="Show citation")
    p.add_argument('--verbose', action='store_true', default=False, help='Write more stuff')
    p.add_argument('-v', '--version', action='version',
                   version='phylophlan_strain_finder.py version {} ({})'.format(__version__, __date__),
                   help="Prints the current phylophlan_strain_finder.py version and exit")

    return p.parse_args()


def check_params(args, verbose=False):
    if verbose:
        info('Checking for parameters...\n')

    if (not os.path.isfile(args.input)):
        error('input file {} does not exist'.format(args.input), exit=True)

    if (not os.path.isfile(args.mutation_rates)):
        error('mutation_rates file {} does not exist'.format(args.map), exit=True)

    if args.output is None:
        args.output = sys.stdout
    else:
        if(os.path.dirname(args.output)):
            if not os.path.exists(os.path.dirname(args.output)):
                error('output path does not exists: "{}"'.format(args.output), exit=True)

        if not os.path.basename(args.output):
            old = args.output
            out_ext = OUTPUT_EXTENSIONS[args.separator] if args.separator in OUTPUT_EXTENSIONS else '.txt'
            args.output = os.path.join(args.output, 'phylophlan_strain_finder.{}'.format(out_ext))
            info('No output filename specified "{}", writing output to "{}"'.format(old, args.output))

        if (not args.overwrite) and os.path.isfile(args.output):
            old = args.output
            args.output = '{}_{}{}'.format(os.path.splitext(args.output)[0],
                                           datetime.datetime.today().strftime('%Y%m%d%H%M%S'),
                                           os.path.splitext(args.output)[1])
            info('Output file "{}" exists, new output filename is "{}"\n'.format(old, args.output))

    if args.phylo_thr < 0.0:
        error('phylo_thr should be a positive number', exit=True)

    if args.mutrate_thr < 0.0:
        error('mutrate_thr should be a positive number', exit=True)

    if verbose:
        info('Arguments: {}\n'.format(vars(args)))


def get_parent(tree, child_clade):
    node_path = tree.get_path(child_clade)

    if len(node_path) > 1:
        return node_path[-2]
    else:
        return child_clade


def check_thr(p, l, tree, md, p_thr, m_thr, verbose=False):
    if p == l:
        if verbose:
            info('Root reached, return {} as root of the subtree\n'.format(l))

        return l

    if any(((tree.distance(other_children, l_children)) / tree.total_branch_length()) > p_thr
           for other_children in p.get_terminals()
           for l_children in l.get_terminals()):
        if verbose:
            info('Not every leaf under {} respects the phylogenetic threshold,\n'
                 'return {} as root of the subtree\n'.format(p, l))

        return l

    else:
        sons = [s.name for s in p.get_terminals()]
        tup = list(itertools.combinations(sons, 2))
        tup = [(sorted(t)) for t in tup]
        if any(md[(t[0], t[1])] > m_thr for t in tup):
            if verbose:
                info('Not every leaf under "{}" respects the mutation_rates threshold, return "{}"" as root of the subtree\n'.format(p, l))
            return l
        else:
            return check_thr(get_parent(tree, p), p, tree, md, p_thr, m_thr, verbose=verbose)


def phylophlan_strain_finder():
    args = read_params()

    if args.verbose:
        info('phylophlan_strain_finder.py version {} ({})\n'.format(__version__, __date__))
        info('Command line: {}\n\n'.format(' '.join(sys.argv)), init_new_line=True)

    check_params(args, args.verbose)
    tree = Phylo.read(args.input, args.tree_format)

    mut_rates = pd.read_csv(args.mutation_rates, sep='\t', header=0, index_col=0)

    if args.verbose:
        info('Reading mutation_rates table...\n')

    mydict = dict([((r, c), float(mut_rates.at[r, c]))
                   for ir, r in enumerate(mut_rates.index)
                   for ic, c in enumerate(mut_rates.columns) if ir < ic])
    mydict.update(dict([((c, r), float(mut_rates.at[r, c]))
                        for ir, r in enumerate(mut_rates.index)
                        for ic, c in enumerate(mut_rates.columns) if (ir < ic) and (c, r) not in mydict]))
    tested_valid = []

    for l in tree.get_terminals():
        if any(l in x.get_terminals() for x in tested_valid):
            continue

        r = (check_thr(get_parent(tree, l), l, tree, mydict, args.phylo_thr, args.mutrate_thr, args.verbose))
        tested_valid.append(r)

    if args.verbose:
        info('Creating output...\n')

    f = args.output

    if isinstance(args.output, str):
        f = open(args.output, 'w')

    print('#phylogenetic_threshold{}{}'.format(args.separator, args.phylo_thr), file=f)
    print('#mutation_rate_threshold{}{}'.format(args.separator, args.mutrate_thr), file=f)
    print('#total_branch_length{}{}'.format(args.separator, tree.total_branch_length()), file=f)
    print(args.separator.join(['#subtree', 'min_dist', 'mean_dist', 'max_dist', 'min_mut',
                               'mean_mut', 'max_mut', 'distances', 'mutation_rates']), file=f)

    for test in tested_valid:
        subtree = Phylo.Newick.Tree(test).format(args.tree_format).strip()
        sons = [s for s in test.get_terminals()]

        if len(sons) == 1:
            continue

        tup = list(itertools.combinations(sons, 2))
        m_rate = []
        distances = []
        m_min = d_min = tree.total_branch_length()
        m_max = m_mean = d_max = d_mean = count = 0

        for t in tup:
            m = mydict[(t[0].name, t[1].name)]
            d = tree.distance(t[0], t[1])
            distances.append(t[0].name + ',' + t[1].name + ':' + str(d))
            m_rate.append(t[0].name + ',' + t[1].name + ':' + str(m))
            m_min = min(m_min, m)
            m_max = max(m_max, m)
            m_mean = m_mean + m
            d_min = min(d_min, d)
            d_max = max(d_max, d)
            d_mean = d_mean + d
            count = count + 1

        d_mean = d_mean / count
        m_mean = m_mean / count

        print(args.separator.join([subtree, str(d_min), str(d_mean), str(d_max), str(m_min), str(m_mean), str(m_max),
                                   '|'.join(distances), '|'.join(m_rate)]), file=f)

    if isinstance(args.output, str):
        if not f.closed:
            f.close()


if __name__ == '__main__':
    t0 = time.time()
    phylophlan_strain_finder()
    t1 = time.time()
    info('Total elapsed time {}s\n'.format(int(t1 - t0)), init_new_line=True)
    sys.exit(0)



#!/usr/bin/env python


__author__ = ('Francesco Asnicar (f.asnicar@unitn.it), '
              'Francesco Beghini (francesco.beghini@unitn.it), '
              'Claudia Mengoni (claudia.mengoni@studenti.unitn.it),'
              'Mattia Bolzan (mattia.bolzan@unitn.it), '
              'Nicola Segata (nicola.segata@unitn.it)')
__version__ = '3.0.19'
__date__ = '3 November 2020'


import os
import sys
import stat
import time
import argparse as ap
import configparser as cp
from distutils.spawn import find_executable


if sys.version_info[0] < 3:
    raise Exception("PhyloPhlAn {} requires Python 3, your current Python version is {}.{}.{}"
                    .format(__version__, sys.version_info[0], sys.version_info[1], sys.version_info[2]))

DB_TYPE_CHOICES = ['n', 'a']
DB_DNA_CHOICES = ['makeblastdb']
DB_AA_CHOICES = ['usearch', 'diamond']
MAP_DNA_CHOICES = ['blastn', 'tblastn', 'diamond']
MAP_AA_CHOICES = ['usearch', 'diamond']
MSA_CHOICES = ['muscle', 'mafft', 'opal', 'upp']
TRIM_CHOICES = ['trimal']
GENE_TREE1_CHOICES = ['fasttree', 'raxml', 'iqtree']
GENE_TREE2_CHOICES = ['raxml']
TREE1_CHOICES = ['fasttree', 'raxml', 'iqtree', 'astral', 'astrid']
TREE2_CHOICES = ['raxml']


def info(s, init_new_line=False, exit=False, exit_value=0):
    if init_new_line:
        sys.stdout.write('\n')

    sys.stdout.write('{}'.format(s))
    sys.stdout.flush()

    if exit:
        sys.exit(exit_value)


def error(s, init_new_line=False, exit=False, exit_value=1):
    if init_new_line:
        sys.stderr.write('\n')

    sys.stderr.write('[e] {}\n'.format(s))
    sys.stderr.flush()

    if exit:
        sys.exit(exit_value)


def read_params():
    p = ap.ArgumentParser(description=("The phylophlan_write_config_file.py script generates a configuration file to be used with the "
                                       "phylophlan.py script. It implements some standard parameters for the software integrated, but if "
                                       "needed, the parameters of the selected software can be added/modified/removed by editing the "
                                       "generated configuration file using a text editor"),
                          formatter_class=ap.ArgumentDefaultsHelpFormatter)

    p.add_argument('-o', '--output', type=str, required=True,
                   help="Specify the output file where to write the configurations")
    p.add_argument('-d', '--db_type', required=True, choices=DB_TYPE_CHOICES,
                   help='Specify the type of the database, where "n" stands for nucleotides and "a" for amino acids')

    p_db = p.add_mutually_exclusive_group(required=True)
    p_db.add_argument('--db_dna', default=None, choices=DB_DNA_CHOICES,
                      help='Add the "db_dna" section of the selected software that will be used for building the indexed database')
    p_db.add_argument('--db_aa', default=None, choices=DB_AA_CHOICES,
                      help='Add the "db_aa" section of the selected software that will be used for building the indexed database')

    p.add_argument('--map_dna', default=None, choices=MAP_DNA_CHOICES,
                   help='Add the "map_dna" section of the selected software that will be used for mapping the database against '
                        'the input genomes')
    p.add_argument('--map_aa', default=None, choices=MAP_AA_CHOICES,
                   help='Add the "map_aa" section of the selected software that will be used for mapping the database against '
                        'the input proteomes')
    p.add_argument('--msa', required=True, default=None, choices=MSA_CHOICES,
                   help='Add the "msa" section of the selected software that will be used for producing the MSAs')
    p.add_argument('--trim', default=None, choices=TRIM_CHOICES,
                   help='Add the "trim" section of the selected software that will be used for the gappy regions removal of the MSAs')
    p.add_argument('--gene_tree1', default=None, choices=GENE_TREE1_CHOICES,
                   help='Add the "gene_tree1" section of the selected software that will be used for building the phylogenies for '
                        'the markers in the database')
    p.add_argument('--gene_tree2', default=None, choices=GENE_TREE2_CHOICES,
                   help='Add the "gene_tree2" section of the selected software that will be used for refining '
                        'the phylogenies previously built with what specified in the "gene_tree1" section')
    p.add_argument('--tree1', required=True, default=None, choices=TREE1_CHOICES,
                   help='Add the "tree1" section of the selected software that will be used for building the first phylogeny')
    p.add_argument('--tree2', default=None, choices=TREE2_CHOICES,
                   help=('Add the "tree2" section of the selected software that will be used for refining the '
                         'phylogeny previously built with what specified in the "tree1" section'))
    p.add_argument('-a', '--absolute_path', action='store_true', default=False,
                   help="Write the absolute path to the executable instead of the executable name as found in the system path environment")
    p.add_argument('--force_nucleotides', default=None, action='store_true',
                   help=('If specified sets parameters for phylogenetic analysis software so that they use '
                         'nucleotide sequences, even in the case of a database of amino acids'))
    p.add_argument('--overwrite', action='store_true', default=False, help="Overwrite output file if it exists")
    p.add_argument('--citation', action='version',
                   version=('Asnicar, F., Thomas, A.M., Beghini, F. et al. '
                            'Precise phylogenetic analysis of microbial isolates and genomes from metagenomes using PhyloPhlAn 3.0. '
                            'Nat Commun 11, 2500 (2020). '
                            'https://doi.org/10.1038/s41467-020-16366-7'),
                   help="Show citation")
    p.add_argument('--verbose', action='store_true', default=False, help="Prints more stuff")
    p.add_argument('-v', '--version', action='version',
                   version='phylophlan_write_config_file.py version {} ({})'.format(__version__, __date__),
                   help="Prints the current phylophlan_write_config_file.py version and exit")
    return p.parse_args()


def check_params(args, verbose=False):
    if (not args.map_dna) and (not args.map_aa):
        error('at least one of --map_dna and --map_aa must be specified', exit=True)

    if (os.path.isfile(args.output)) and (not args.overwrite):
        error('output file {} already exists, delete it or specify --overwrite'.format(args.output), exit=True)

    if (args.db_type == 'n') and (args.map_dna == 'diamond'):
        error('Incompatible choices.\n    Please, select "blastn" instead in --map_dna, as there is no "diamond" '
              'option for aligning against a nucleotide reference database', exit=True)

    if verbose:
        info('Arguments: {}\n'.format(vars(args)))


def find_executable_wrapper(exe, absolute=False, rollback=False):
    find1 = find_executable(exe)
    find2 = find_executable(rollback) if rollback else None

    if (find1 is None) and (find2 is None):
        error('could not find "{}" ("{}") executable in your PATH environment variable'.format(exe, rollback), exit=True)

    return (find1 if find1 else find2, find1 is None)


# AVAILABLE OPTIONS:
# program_name: name of the executable to use
# params: params to use
# threads: specify the option to use to pass the number of threads
# input: specify the option to use for the input file
# database: specify the option to use for setting the database
# output_path: specify the to option to use to set the path of the folder that will
#              contains the output file
# output: specify the option to use for the output file
# version: specify the option to use to get the version of the software, used to verify
#          the software installation
# command_line: specify the command line to generate with the position of each argument,
#               '<' and '>' can be used to specify input/output redirection, respectively
# environment: specify variables and their values to be defined in the environment,
#              syntax VARIABLE1=VALUE1,VARIABLE2=VALUE2,...


def phylophlan_write_config_file():
    args = read_params()

    if args.verbose:
        info('phylophlan_write_config_file.py version {} ({})\n'.format(__version__, __date__))
        info('Command line: {}\n\n'.format(' '.join(sys.argv)), init_new_line=True)

    check_params(args, verbose=args.verbose)
    progs = {}

    # setting the program for building the nucleotides DB
    if args.db_dna:
        if 'makeblastdb' in args.db_dna:
            exe, _ = find_executable_wrapper('makeblastdb', absolute=args.absolute_path)
            db_dna = {'program_name': exe,
                      'params': '-parse_seqids -dbtype nucl',
                      'input': '-in',
                      'output': '-out',
                      'version': '-version',
                      'command_line': '#program_name# #params# #input# #output#'}

        progs['db_dna'] = db_dna

    # setting the program for building the amino acids DB
    if args.db_aa:
        if 'usearch' in args.db_aa:
            exe, _ = find_executable_wrapper('usearch', absolute=args.absolute_path)
            db_aa = {'program_name': exe,
                     'params': '-quiet',
                     'input': '-makeudb_ublast',
                     'output': '-output',
                     'version': '-version',
                     'command_line': '#program_name# #params# #input# #output#'}
        elif 'diamond' in args.db_aa:
            exe, _ = find_executable_wrapper('diamond', absolute=args.absolute_path)
            db_aa = {'program_name': exe,
                     'params': 'makedb',
                     'threads': '--threads',
                     'input': '--in',
                     'output': '--db',
                     'version': 'version',
                     'command_line': '#program_name# #params# #threads# #input# #output#'}

        progs['db_aa'] = db_aa

    # setting the software for mapping the genomes (dna)
    if args.map_dna:
        if 'blastn' in args.map_dna:
            exe, _ = find_executable_wrapper('blastn', absolute=args.absolute_path)
            map_dna = {'program_name': exe,
                       'params': '-outfmt 6 -evalue 0.1 -max_target_seqs 1000000 -perc_identity 75',
                       'input': '-query',
                       'database': '-db',
                       'output': '-out',
                       'version': '-version',
                       'command_line': '#program_name# #params# #input# #database# #output#'}
        elif 'tblastn' in args.map_dna:
            exe, _ = find_executable_wrapper('tblastn', absolute=args.absolute_path)
            map_dna = {'program_name': exe,
                       'params': ('-outfmt "6 saccver qaccver pident length mismatch gapopen sstart send qstart '
                                  'qend evalue bitscore" -evalue 1e-50 -max_target_seqs 1000000 -perc_identity 50'),
                       'input': '-subject',
                       'database': '-query',
                       'output': '-out',
                       'version': '-version',
                       'command_line': '#program_name# #params# #input# #database# #output#'}
        elif 'diamond' in args.map_dna:
            exe, _ = find_executable_wrapper('diamond', absolute=args.absolute_path)
            map_dna = {'program_name': exe,
                       'params': 'blastx --quiet --threads 1 --outfmt 6 --more-sensitive --id 50 --max-hsps 35 -k 0',
                       'input': '--query',
                       'database': '--db',
                       'output': '--out',
                       'version': 'version',
                       'command_line': '#program_name# #params# #input# #database# #output#'}

        progs['map_dna'] = map_dna

    # setting the software for mapping the proteomes (aa)
    if args.map_aa:
        if 'usearch' in args.map_aa:
            exe, _ = find_executable_wrapper('usearch', absolute=args.absolute_path)
            map_aa = {'program_name': exe,
                      'params': '-quiet -evalue 1e-10 -maxaccepts 8 -maxrejects 32',
                      'threads': '-threads',
                      'input': '-ublast',
                      'database': '-db',
                      'output': '-blast6out',
                      'version': '-version',
                      'command_line': '#program_name# #params# #threads# #input# #database# #output#'}
        elif 'diamond' in args.map_aa:
            exe, _ = find_executable_wrapper('diamond', absolute=args.absolute_path)
            map_aa = {'program_name': exe,
                      'params': 'blastp --quiet --threads 1 --outfmt 6 --more-sensitive --id 50 --max-hsps 35 -k 0',
                      'input': '--query',
                      'database': '--db',
                      'output': '--out',
                      'version': 'version',
                      'command_line': '#program_name# #params# #input# #database# #output#'}

        progs['map_aa'] = map_aa

    # setting the MSA software
    if 'muscle' in args.msa:
        exe, _ = find_executable_wrapper('muscle', absolute=args.absolute_path)
        msa = {'program_name': exe,
               'params': '-quiet -maxiters 2',
               'input': '-in',
               'output': '-out',
               'version': '-version',
               'command_line': '#program_name# #params# #input# #output#'}
    elif 'mafft' in args.msa:
        exe, _ = find_executable_wrapper('mafft', absolute=args.absolute_path)
        msa = {'program_name': exe,
               'params': '--quiet --anysymbol --thread 1 --auto',
               'version': '--version',
               'command_line': '#program_name# #params# #input# > #output#'}

        for fld in ['/local-storage', '/tmp']:
            if os.path.isdir(fld) and os.access(fld, os.W_OK) and os.access(fld, os.X_OK):
                if 'environment' in msa:
                    break

                msa['environment'] = 'TMPDIR={}'.format(fld)
                break
    elif 'opal' in args.msa:
        exe, _ = find_executable_wrapper('opal', absolute=args.absolute_path)
        msa = {'program_name': exe,
               'input': '--in',
               'output': '--out',
               'params': '--quiet',
               'command_line': '#program_name# #params# #input# #output#'}

        if (args.db_type == 'a' and (not args.force_nucleotides)):
            msa['params'] += ' --protein'
    elif 'upp' in args.msa:
        exe, _ = find_executable_wrapper('upp', rollback='run-upp.sh', absolute=args.absolute_path)
        msa = {'program_name': exe,
               'params': '-x 1 -M -1 -T 0.66 -B 999999999',
               'input': '-s',
               'output': '-o',
               'output_path': '-d',
               'version': '--version',
               'command_line': '#program_name# #params# #input# #output_path# #output#'}

        if (args.db_type == 'n' or args.force_nucleotides):
            msa['params'] += ' -m dna',
        elif args.db_type == 'a':
            msa['model'] += ' -m amino'

    progs['msa'] = msa

    # setting the trimming software
    if args.trim:
        if 'trimal' in args.trim:
            exe, _ = find_executable_wrapper('trimal', absolute=args.absolute_path)
            trim = {'program_name': exe,
                    'params': '-gappyout',
                    'input': '-in',
                    'output': '-out',
                    'version': '--version',
                    'command_line': '#program_name# #params# #input# #output#'}

        progs['trim'] = trim

    # setting gene_tree1
    if args.gene_tree1:
        if 'fasttree' in args.gene_tree1:
            exe, _ = find_executable_wrapper('FastTree', rollback='fasttree', absolute=args.absolute_path)
            gene_tree1 = {'program_name': exe,
                          'params': '-quiet -pseudo -spr 4 -mlacc 2 -slownni -fastest -no2nd -mlnni 4',
                          'output': '-out',
                          'command_line': '#program_name# #params# #output# #input#'}

            if (args.db_type == 'n' or args.force_nucleotides):
                gene_tree1['params'] += ' -gtr -nt'
            elif args.db_type == 'a':
                gene_tree1['params'] += ' -lg'
        elif 'raxml' in args.gene_tree1:
            exe, _ = find_executable_wrapper('raxmlHPC', rollback='raxml', absolute=args.absolute_path)
            gene_tree1 = {'program_name': exe,
                          'params': '-p 1989',
                          'input': '-s',
                          'output_path': '-w',
                          'output': '-n',
                          'version': '-v'}

            if (args.db_type == 'n' or args.force_nucleotides):
                gene_tree1['params'] += ' -m GTRCAT',
                gene_tree1['command_line'] = '#program_name# #params# #output_path# #input# #output#'
            elif args.db_type == 'a':
                gene_tree1['model'] = '-m'
                gene_tree1['command_line'] = '#program_name# #model# #params# #output_path# #input# #output#'
        elif 'iqtree' in args.gene_tree1:
            exe, _ = find_executable_wrapper('iqtree', absolute=args.absolute_path)
            gene_tree1 = {'program_name': exe,
                          'params': '-quiet -nt 1',
                          'input': '-s',
                          'output': '-pre',
                          'version': '-version',
                          'command_line': '#program_name# #params# #input# #output#'}

            if (args.db_type == 'n' or args.force_nucleotides):
                gene_tree1['params'] += ' -m GTR',
            elif args.db_type == 'a':
                gene_tree1['params'] = ' -m LG'

        progs['gene_tree1'] = gene_tree1

    # setting gene_tree2
    if args.gene_tree2:
        if 'raxml' in args.gene_tree2:
            exe, _ = find_executable_wrapper('raxmlHPC', rollback='raxml', absolute=args.absolute_path)
            gene_tree2 = {'program_name': exe,
                          'params': '-p 1989',
                          'database': '-t',  # starting tree
                          'input': '-s',
                          'output_path': '-w',
                          'output': '-n',
                          'version': '-v'}

            if (args.db_type == 'n' or args.force_nucleotides):
                gene_tree2['params'] += '-p 1989 -m GTRCAT'
                gene_tree2['command_line'] = '#program_name# #params# #database# #output_path# #input# #output#'
            elif args.db_type == 'a':
                gene_tree2['model'] = '-m'
                gene_tree2['command_line'] = '#program_name# #model# #params# #database# #output_path# #input# #output#'

        progs['gene_tree2'] = gene_tree2

    # setting tree1
    if 'astral' in args.tree1:
        tree1 = {'program_name': 'java -jar astral-4.11.1/astral.4.11.1.jar',
                 'input': '-i',
                 'output': '-o',
                 'version': '-i astral-4.11.1/test_data/song_mammals.424.gene.tre',
                 'command_line': '#program_name# #input# #output#'}
        info('[w] using a template for ASTRAL as it should be manually installed from https://github.com/smirarab/ASTRAL and this config file should be manually updated\n')
    if 'astrid' in args.tree1:
        exe, _ = find_executable_wrapper('ASTRID', absolute=args.absolute_path)
        tree1 = {'program_name': exe,
                 'input': '-i',
                 'params': '-m auto',
                 'output': '-o',
                 'version': '--help',
                 'command_line': '#program_name# #input# #params# #output#'}
    elif 'fasttree' in args.tree1:
        exe, rb = find_executable_wrapper('FastTreeMP', rollback='fasttree', absolute=args.absolute_path)
        tree1 = {'program_name': exe,
                 'params': '-quiet -pseudo -spr 4 -mlacc 2 -slownni -fastest -no2nd -mlnni 4',
                 'output': '-out',
                 'command_line': '#program_name# #params# #output# #input#'}

        if not rb:
            tree1['environment'] = 'OMP_NUM_THREADS=3'

        if (args.db_type == 'n' or args.force_nucleotides):
            tree1['params'] += ' -gtr -nt'
        elif args.db_type == 'a':
            tree1['params'] += ' -lg'
    elif 'raxml' in args.tree1:
        exe, rb = find_executable_wrapper('raxmlHPC-PTHREADS-SSE3', rollback='raxml', absolute=args.absolute_path)
        tree1 = {'program_name': exe,
                 'params': '-p 1989',
                 'input': '-s',
                 'output_path': '-w',
                 'output': '-n',
                 'version': '-v',
                 'command_line': '#program_name# #params# #threads# #output_path# #input# #output#'}

        if not rb:
            tree1['threads'] = '-T'

        if (args.db_type == 'n' or args.force_nucleotides):
            tree1['params'] += ' -m GTRCAT'
        elif args.db_type == 'a':
            tree1['params'] += ' -m PROTCATLG'

    elif 'iqtree' in args.tree1:
        exe, _ = find_executable_wrapper('iqtree', absolute=args.absolute_path)
        tree1 = {'program_name': exe,
                 'params': '-quiet -nt AUTO',
                 'input': '-s',
                 'output': '-pre',
                 'version': '-version',
                 'command_line': '#program_name# #params# #input# #output#'}

        if (args.db_type == 'n' or args.force_nucleotides):
            tree1['params'] += ' -m GTR'
        elif args.db_type == 'a':
            tree1['params'] += ' -m LG'

    progs['tree1'] = tree1

    # setting tree2
    if args.tree2:
        if 'raxml' in args.tree2:
            exe, rb = find_executable_wrapper('raxmlHPC-PTHREADS-SSE3', rollback='raxml', absolute=args.absolute_path)
            tree2 = {'program_name': exe,
                     'params': '-p 1989',
                     'database': '-t',  # starting tree
                     'input': '-s',
                     'output_path': '-w',
                     'output': '-n',
                     'version': '-v',
                     'command_line': '#program_name# #params# #threads# #database# #output_path# #input# #output#'}

            if not rb:
                tree2['threads'] = '-T'

            if (args.db_type == 'n' or args.force_nucleotides):
                tree2['params'] += ' -m GTRCAT'
            elif args.db_type == 'a':
                tree2['params'] += ' -m PROTCATLG'

        progs['tree2'] = tree2

    config = cp.ConfigParser()

    for prog, options in progs.items():
        config[prog] = {}

        for option, value in options.items():
            config[prog][option] = value

    if (os.path.isfile(args.output)) and args.overwrite and args.verbose:
        info('Output file "{}" will be overwritten\n'.format(args.output))

    with open(args.output, 'w') as f:
        config.write(f)


if __name__ == '__main__':
    t0 = time.time()
    phylophlan_write_config_file()
    t1 = time.time()
    info('Total elapsed time {}s\n'.format(int(t1 - t0)), init_new_line=True)
    sys.exit(0)
